"""Settings for HFMC daemon service."""

from aiohttp import ClientTimeout

ApiType = str

API_PREFIX: ApiType = "/hfmc_api/{service}"

API_PEERS_PROBE: ApiType = API_PREFIX.format(service="peers/ping")

API_DAEMON_RUNNING: ApiType = API_PREFIX.format(service="daemon/status")
API_DAEMON_STOP: ApiType = API_PREFIX.format(service="daemon/stop")
API_DAEMON_PEERS_ALIVE: ApiType = API_PREFIX.format(
    service="daemon/peers_alive",
)
API_DAEMON_PEERS_CHANGE: ApiType = API_PREFIX.format(
    service="daemon/peers_change",
)

API_FETCH_FILE_CLIENT: ApiType = "/{repo}/resolve/{revision}/{file_name}"
API_FETCH_FILE_DAEMON: ApiType = "/{user}/{model}/resolve/{revision}/{file_name:.*}"
API_FETCH_REPO_FILE_LIST: ApiType = API_PREFIX.format(
    service="fetch/repo_file_list/{user}/{model}/{revision}"
)


# timeout in sec
TIMEOUT_PEERS = ClientTimeout(total=10)
TIMEOUT_DAEMON = ClientTimeout(total=2)
TIMEOUT_FETCH = ClientTimeout(total=30)



"""Context Peer definition for HFMC."""

from dataclasses import dataclass, field


@dataclass(order=True, unsafe_hash=True)
class Peer:
    """Peer definition for HFMC."""

    ip: str = field(hash=True)
    port: int = field(hash=True)
    alive: bool = field(compare=False, default=False)
    epoch: int = field(compare=False, default=0)



"""Utils for yaml files."""

from pathlib import Path

import yaml
from pydantic import BaseModel


def yaml_load(file_path: Path) -> dict:
    """Load a yaml file."""
    return yaml.safe_load(file_path.read_text())


def yaml_dump(model: BaseModel) -> str:
    """Dump a model to a yaml string."""
    return yaml.dump(model.model_dump())



"""Define HFMC context."""

from __future__ import annotations

from dataclasses import dataclass, field
from pathlib import Path
from typing import TYPE_CHECKING, List

from hfmc.common.peer import Peer

if TYPE_CHECKING:
    from hfmc.config.hfmc_config import HfmcConfig
    from hfmc.daemon.prober import PeerProber


@dataclass()
class HfmcContext:
    """HFMC context."""

    # properties
    port: int = field()
    model_dir: Path = field()
    etag_dir: Path = field()
    log_dir: Path = field()
    repo_files_dir: Path = field()
    peers: List[Peer] = field()
    peer_prober: PeerProber | None = field(
        default=None,
        init=False,
        repr=False,
    )

    # global context reference
    _instance: HfmcContext | None = field(
        default=None,
        init=False,
        repr=False,
    )

    @classmethod
    def init_with_config(cls, config: HfmcConfig) -> HfmcContext:
        """Create HFMC context from configuration."""
        cls._instance = cls(
            port=config.daemon_port,
            model_dir=Path(config.cache_dir) / "models",
            etag_dir=Path(config.cache_dir) / "etags",
            log_dir=Path(config.cache_dir) / "logs",
            repo_files_dir=Path(config.cache_dir) / "repo_files",
            peers=[Peer(ip=p.ip, port=p.port) for p in config.peers],
        )
        if not cls.get_model_dir().exists():
            cls.get_model_dir().mkdir(parents=True, exist_ok=True)
        if not cls.get_etag_dir().exists():
            cls.get_etag_dir().mkdir(parents=True, exist_ok=True)
        if not cls.get_log_dir().exists():
            cls.get_log_dir().mkdir(parents=True, exist_ok=True)
        if not cls.get_repo_files_dir().exists():
            cls.get_repo_files_dir().mkdir(parents=True, exist_ok=True)
        return cls._instance

    @classmethod
    def get_port(cls) -> int:
        """Get port."""
        if not cls._instance:
            raise ValueError
        return cls._instance.port

    @classmethod
    def get_model_dir(cls) -> Path:
        """Get model dir."""
        if not cls._instance:
            raise ValueError
        return cls._instance.model_dir

    @classmethod
    def get_model_dir_str(cls) -> str:
        """Get model dir in str."""
        return str(cls.get_model_dir())

    @classmethod
    def get_etag_dir(cls) -> Path:
        """Get etag dir."""
        if not cls._instance:
            raise ValueError
        return cls._instance.etag_dir

    @classmethod
    def get_log_dir(cls) -> Path:
        """Get log dir."""
        if not cls._instance:
            raise ValueError
        return cls._instance.log_dir

    @classmethod
    def get_repo_files_dir(cls) -> Path:
        """Get repo file list dir."""
        if not cls._instance:
            raise ValueError
        return cls._instance.repo_files_dir

    @classmethod
    def get_peers(cls) -> List[Peer]:
        """Get peers."""
        if not cls._instance:
            raise ValueError
        return cls._instance.peers

    @classmethod
    def update_peers(
        cls,
        conf: HfmcConfig,
        old_peers: List[Peer],
    ) -> List[Peer]:
        """Load peers from config and then update their states."""
        if not cls._instance:
            raise ValueError

        new_peers = [Peer(ip=p.ip, port=p.port) for p in conf.peers]
        peer_map = {p: p for p in new_peers}

        for peer in old_peers:
            if peer in peer_map:  # peer match by ip and port
                peer_map[peer].alive = peer.alive
                peer_map[peer].epoch = peer.epoch

        cls._instance.peers = list(peer_map.values())

        return cls._instance.peers

    @classmethod
    def get_daemon(cls) -> Peer:
        """Get local daemon."""
        if not cls._instance:
            raise ValueError
        return Peer(ip="127.0.0.1", port=cls._instance.port)

    @classmethod
    def set_peer_prober(cls, peer_prober: PeerProber) -> None:
        """Set peer prober."""
        if not cls._instance:
            raise ValueError
        cls._instance.peer_prober = peer_prober

    @classmethod
    def get_peer_prober(cls) -> PeerProber:
        """Get peer prober."""
        if not cls._instance:
            raise ValueError
        if not cls._instance.peer_prober:
            raise ValueError
        return cls._instance.peer_prober



"""Initialize, load, and save configuration settings."""

from __future__ import annotations

from typing import TypeVar, cast

from hfmc.utils.yaml import yaml_dump, yaml_load

from .hfmc_config import (
    CONFIG_DIR,
    CONFIG_FILE,
    HfmcConfig,
)
from .hfmc_config import (
    HfmcConfigOption as ConfOpt,
)

DEFAULT_CONFIG = HfmcConfig()


def init_config() -> None:
    """Initialize HFMC configuration."""
    if not CONFIG_DIR.exists():
        CONFIG_DIR.mkdir(parents=True, exist_ok=True)

    if not CONFIG_FILE.exists():
        CONFIG_FILE.touch()

    # create yaml files based on the configuration settings
    CONFIG_FILE.write_text(yaml_dump(DEFAULT_CONFIG))


def load_config() -> HfmcConfig:
    """Load HFMC configuration."""
    if not CONFIG_FILE.exists():
        init_config()

    conf_dict = yaml_load(CONFIG_FILE)
    return HfmcConfig.model_validate(conf_dict)


def save_config(config: HfmcConfig) -> None:
    """Save HFMC configuration."""
    if not CONFIG_FILE.exists():
        init_config()

    CONFIG_FILE.write_text(yaml_dump(config))


def get_config_yaml() -> str:
    """Get HFMC configuration in yaml format."""
    return CONFIG_FILE.read_text()


T = TypeVar("T")


def get_config(opt: ConfOpt, _: type[T]) -> T:
    """Get a specific configuration option."""
    config = load_config()
    return cast(T, getattr(config, opt.value))


def set_config(opt: ConfOpt, value: T, _: type[T]) -> T:
    """Set a specific configuration option."""
    config = load_config()
    setattr(config, opt.value, cast(T, value))
    save_config(config)
    return value


def reset_config(opt: ConfOpt, conf_type: type[T]) -> T:
    """Reset a specific configuration option."""
    value = cast(T, getattr(DEFAULT_CONFIG, opt.value))
    return set_config(opt, value, conf_type)



"""Configuration related commands."""

import logging
from argparse import Namespace

from hfmc.config import config_manager
from hfmc.config.hfmc_config import HfmcConfigOption

logger = logging.getLogger(__name__)


def _configure_cache(args: Namespace) -> None:
    if args.conf_cache_command == "set":
        conf = config_manager.set_config(
            HfmcConfigOption.CACHE,
            args.path,
            str,
        )
        logger.info("Set HFMC cache path: %s", conf)
    elif args.conf_cache_command == "get":
        conf = config_manager.get_config(HfmcConfigOption.CACHE, str)
        logger.info("HFMC cache path: %s", conf)
    elif args.conf_cache_command == "reset":
        conf = config_manager.reset_config(HfmcConfigOption.CACHE, str)
        logger.info("Reset HFMC cache path: %s", conf)


def _configure_port(args: Namespace) -> None:
    if args.conf_port_command == "set":
        conf = config_manager.set_config(
            HfmcConfigOption.PORT,
            args.port,
            str,
        )
        logger.info("Set HFMC port: %s", conf)
    elif args.conf_port_command == "get":
        conf = config_manager.get_config(HfmcConfigOption.PORT, str)
        logger.info("HFMC port: %s", conf)
    elif args.conf_port_command == "reset":
        conf = config_manager.reset_config(HfmcConfigOption.PORT, str)
        logger.info("Reset HFMC port: %s", conf)


def _show_config() -> None:
    content = config_manager.get_config_yaml()
    logger.info(content)


async def exec_cmd(args: Namespace) -> None:
    """Execute command."""
    if args.conf_command == "cache":
        _configure_cache(args)
    elif args.conf_command == "port":
        _configure_port(args)
    elif args.conf_command == "show":
        _show_config()
    else:
        raise NotImplementedError



"""Daemon client for connecting with (self or other) Daemons."""

from __future__ import annotations

import asyncio
import logging
import time
from contextlib import asynccontextmanager
from typing import AsyncContextManager, AsyncIterator, List

import aiohttp
from huggingface_hub import (  # type: ignore[import-untyped]
    get_hf_file_metadata,
    hf_hub_url,
)

from hfmc.common.context import HfmcContext
from hfmc.common.api_settings import (
    API_DAEMON_PEERS_ALIVE,
    API_DAEMON_PEERS_CHANGE,
    API_DAEMON_RUNNING,
    API_DAEMON_STOP,
    API_FETCH_FILE_CLIENT,
    API_FETCH_REPO_FILE_LIST,
    API_PEERS_PROBE,
    TIMEOUT_DAEMON,
    TIMEOUT_PEERS,
    ApiType,
)
from hfmc.common.peer import Peer
from hfmc.common.repo_files import RepoFileList

logger = logging.getLogger(__name__)


HTTP_STATUS_OK = 200


def _http_session() -> aiohttp.ClientSession:
    return aiohttp.ClientSession()


def _api_url(peer: Peer, api: ApiType) -> str:
    return f"http://{peer.ip}:{peer.port}{api}"


@asynccontextmanager
async def _quiet_request(
    sess: aiohttp.ClientSession,
    req: AsyncContextManager,
) -> AsyncIterator[aiohttp.ClientResponse | None]:
    try:
        async with sess, req as resp:
            yield resp
    except (
        aiohttp.ClientError,
        asyncio.exceptions.TimeoutError,
        TimeoutError,
        ConnectionError,
        RuntimeError,
    ) as e:
        logger.debug("HTTP Error: %s", e)
        yield None


@asynccontextmanager
async def _quiet_get(
    url: str,
    timeout: aiohttp.ClientTimeout,
) -> AsyncIterator[aiohttp.ClientResponse | None]:
    sess = _http_session()
    req = sess.get(url, timeout=timeout)
    async with _quiet_request(sess, req) as resp:
        try:
            yield resp
        except (OSError, ValueError, RuntimeError) as e:
            logger.debug("Failed to get response: %s", e)
            yield None


@asynccontextmanager
async def _quiet_head(
    url: str,
    timeout: aiohttp.ClientTimeout,
) -> AsyncIterator[aiohttp.ClientResponse | None]:
    sess = _http_session()
    req = sess.head(url, timeout=timeout)
    async with _quiet_request(sess, req) as resp:
        try:
            yield resp
        except (OSError, ValueError, RuntimeError) as e:
            logger.debug("Failed to get response: %s", e)
            yield None


async def ping(target: Peer) -> Peer:
    """Ping a peer to check if it is alive."""
    url = _api_url(target, API_PEERS_PROBE)
    async with _quiet_get(url, TIMEOUT_PEERS) as resp:
        target.alive = resp is not None and resp.status == HTTP_STATUS_OK
        target.epoch = int(time.time())
        return target


async def stop_daemon() -> bool:
    """Stop a daemon service."""
    url = _api_url(HfmcContext.get_daemon(), API_DAEMON_STOP)
    async with _quiet_get(url, TIMEOUT_DAEMON) as resp:
        return resp is not None and resp.status == HTTP_STATUS_OK


async def is_daemon_running() -> bool:
    """Check if daemon is running."""
    url = _api_url(HfmcContext.get_daemon(), API_DAEMON_RUNNING)
    async with _quiet_get(url, TIMEOUT_DAEMON) as resp:
        return resp is not None and resp.status == HTTP_STATUS_OK


async def get_alive_peers() -> List[Peer]:
    """Get a list of alive peers."""
    url = _api_url(HfmcContext.get_daemon(), API_DAEMON_PEERS_ALIVE)
    async with _quiet_get(url, TIMEOUT_DAEMON) as resp:
        if not resp:
            return []
        return [Peer(**peer) for peer in await resp.json()]


async def notify_peers_change() -> bool:
    """Notify peers about a change in peer list."""
    url = _api_url(HfmcContext.get_daemon(), API_DAEMON_PEERS_CHANGE)
    async with _quiet_get(url, TIMEOUT_DAEMON) as resp:
        return resp is not None and resp.status == HTTP_STATUS_OK


async def check_file_exist(
    peer: Peer,
    repo_id: str,
    file_name: str,
    revision: str,
) -> tuple[Peer, bool]:
    """Check if the peer has target file."""
    url = _api_url(
        peer,
        API_FETCH_FILE_CLIENT.format(
            repo=repo_id,
            revision=revision,
            file_name=file_name,
        ),
    )
    async with _quiet_head(url, TIMEOUT_PEERS) as resp:
        return (peer, resp is not None and resp.status == HTTP_STATUS_OK)


async def get_file_etag(
    endpoint: str,
    repo_id: str,
    file_name: str,
    revision: str,
) -> str | None:
    """Get the ETag of a file."""
    url = hf_hub_url(
        repo_id=repo_id,
        filename=file_name,
        revision=revision,
        endpoint=endpoint,
    )

    try:
        metadata = get_hf_file_metadata(url)
        if metadata:
            return metadata.etag
    except (OSError, ValueError):
        logger.debug(
            "Failed to get ETag: %s, %s, %s, %s",
            endpoint,
            repo_id,
            file_name,
            revision,
        )
    return None


async def check_repo_exist() -> tuple[Peer, bool]:
    """Check if the peer has target model."""
    raise NotImplementedError


async def get_repo_file_list(
    peer: Peer,
    repo_id: str,
    revision: str,
) -> RepoFileList | None:
    """Load the target model from a peer."""
    user, model = repo_id.strip().split("/")
    url = _api_url(
        peer,
        API_FETCH_REPO_FILE_LIST.format(
            user=user,
            model=model,
            revision=revision,
        ),
    )
    async with _quiet_get(url, TIMEOUT_PEERS) as resp:
        if not resp or resp.status != HTTP_STATUS_OK:
            return None
        return await resp.json()



"""Manage models."""

from __future__ import annotations

import asyncio
import logging
from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Any, Coroutine, List, TypeVar

from huggingface_hub import hf_hub_download  # type: ignore[import-untyped]
from huggingface_hub.utils import GatedRepoError  # type: ignore[import-untyped]

from hfmc.client import http_request as request
from hfmc.common import hf_wrapper
from hfmc.common.context import HfmcContext
from hfmc.common.etag import save_etag
from hfmc.common.repo_files import RepoFileList, load_file_list, save_file_list

if TYPE_CHECKING:
    from pathlib import Path

    from hfmc.common.peer import Peer

logger = logging.getLogger(__name__)


T = TypeVar("T")


async def _safe_gather(
    tasks: List[Coroutine[Any, Any, T]],
) -> List[T]:
    results = await asyncio.gather(*tasks, return_exceptions=True)
    return [r for r in results if not isinstance(r, BaseException)]


async def file_search(
    repo_id: str,
    file_name: str,
    revision: str,
) -> List[Peer]:
    """Check which peers have target file."""
    alives = await request.get_alive_peers()
    tasks = [
        request.check_file_exist(alive, repo_id, file_name, revision)
        for alive in alives
    ]
    results = await _safe_gather(tasks)
    exists = {s[0] for s in results if s[1]}
    return [alive for alive in alives if alive in exists]


async def repo_search() -> List[Peer]:
    """Check which peers have target model."""
    raise NotImplementedError


async def _download_file(
    endpoint: str,
    repo_id: str,
    file_name: str,
    revision: str,
) -> bool:
    try:
        # hf_hub_download will send request to the endpoint
        # on /{user}/{model}/resolve/{revision}/{file_name:.*}
        # daemon server can handle the request and return the file
        _ = hf_hub_download(
            endpoint=endpoint,
            repo_id=repo_id,
            revision=revision,
            filename=file_name,
            cache_dir=HfmcContext.get_model_dir_str(),
        )

        etag = await request.get_file_etag(
            endpoint,
            repo_id,
            file_name,
            revision,
        )
        if not etag:
            return False

        save_etag(etag, repo_id, file_name, revision)
    except GatedRepoError:
        logger.info("Model is gated. Login with `hfmc auth login` first.")
        return False
    except (OSError, ValueError) as e:
        logger.info(f"Failed to download model. ERROR: {e}")
        logger.debug("Download file error", exc_info=e)
        return False
    return True


def _gen_endpoints(peers: List[Peer]) -> List[str]:
    peer_ends = [f"http://{peer.ip}:{peer.port}" for peer in peers]
    site_ends = ["https://hf-mirror.com", "https://huggingface.co"]
    return peer_ends + site_ends


async def file_add(
    repo_id: str,
    file_name: str,
    revision: str,
) -> bool:
    """Download and add model files to HFMC."""
    if hf_wrapper.get_file_info(repo_id, revision, file_name) is not None:
        # file is already downloaded
        return True

    peers = await file_search(repo_id, file_name, revision)
    endpoints = _gen_endpoints(peers)

    for endpoint in endpoints:
        logger.info("Try to add file %s from %s", file_name, endpoint)
        success = await _download_file(endpoint, repo_id, file_name, revision)

        if success:
            return True

    return False


async def _wait_first(
    tasks: List[Coroutine[Any, Any, T | None]],
) -> T | None:
    done, _ = await asyncio.wait(tasks, return_when=asyncio.FIRST_COMPLETED)
    if not done:
        return None
    for task in done:
        result = task.result()
        if result:
            return result
    return None


async def _file_list_from_peers(
    repo_id: str,
    revision: str,
) -> RepoFileList | None:
    alives = await request.get_alive_peers()
    if not alives:
        return None
    tasks = [request.get_repo_file_list(alive, repo_id, revision) for alive in alives]
    return await _wait_first(tasks)


async def _file_list_from_site(
    repo_id: str,
    revision: str,
) -> RepoFileList | None:
    endopints = _gen_endpoints([])
    for endpoint in endopints:
        files = hf_wrapper.get_repo_file_list(endpoint, repo_id, revision)
        if files:
            return files
    return None


async def _get_repo_file_list(
    repo_id: str,
    revision: str,
) -> RepoFileList | None:
    files = load_file_list(repo_id, revision)
    if not files:
        files = await _file_list_from_peers(repo_id, revision)
    if not files:
        files = await _file_list_from_site(repo_id, revision)
    if files:
        save_file_list(repo_id, revision, files)
    return files


async def repo_add(
    repo_id: str,
    revision: str,
) -> bool:
    """Download and add all files in a repo to HFMC."""
    normalized_rev = hf_wrapper.verify_revision(
        repo_id,
        revision,
        _gen_endpoints([]),
    )
    if not normalized_rev:
        logger.error("Failed to verify revision: %s", revision)
        return False

    files = await _get_repo_file_list(repo_id, normalized_rev)
    if not files:
        logger.error("Failed to get file list of %s", repo_id)
        return False

    for file_name in files:
        success = await file_add(repo_id, file_name, normalized_rev)
        if not success:
            logger.error("Failed to add file: %s", file_name)
            return False

    return True


@dataclass
class FileInfo:
    """Info of a model file."""

    file_name: Path = field()
    size_on_disk_str: str = field()
    file_path: Path = field()
    refs: set[str] = field()
    commit_8: str = field()


@dataclass
class RepoInfo:
    """Info of a repo."""

    repo_id: str = field()
    size_str: str = field()
    nb_files: int = field()
    repo_path: Path = field()


def file_list(repo_id: str) -> List[FileInfo]:
    """List files in target repo."""
    files: List[FileInfo] = []

    repo_info = hf_wrapper.get_repo_info(repo_id)
    if not repo_info:
        return files

    for rev in repo_info.revisions:
        for f in rev.files:
            fi = FileInfo(
                f.file_path.relative_to(rev.snapshot_path),
                f.size_on_disk_str,
                f.file_path,
                set(rev.refs),
                rev.commit_hash[:8],
            )
            files.extend([fi])

    return files


def repo_list() -> List[RepoInfo]:
    """List repos in the cache."""
    cache_info = hf_wrapper.get_cache_info()
    return [
        RepoInfo(
            repo.repo_id,
            f"{repo.size_on_disk_str:>12}",
            repo.nb_files,
            repo.repo_path,
        )
        for repo in cache_info.repos
    ]


def _is_relative_to(child: Path, parent: Path) -> bool:
    try:
        _ = child.relative_to(parent)
    except ValueError:
        return False
    return True


def _rm_file(fp: Path, root_path: Path) -> None:
    if not fp.relative_to(root_path):
        logger.debug(
            "Cache structure error: path=%s, root=%s",
            str(fp),
            str(root_path),
        )
        raise ValueError

    # remove target file
    if fp.exists() and fp.is_file():
        fp.unlink()

    # remove parent directories if empty up to root_path
    parent_dir = fp.parent
    while _is_relative_to(parent_dir, root_path):
        if not any(parent_dir.iterdir()):
            parent_dir.rmdir()
            parent_dir = parent_dir.parent
        else:
            break


def _can_delete_blob(
    file_name: str,
    snapshot_path: Path,
    blob_path: Path,
) -> bool:
    """Delete blob only if there is NO symlink pointing to it."""
    if not snapshot_path.exists():
        # delete blob if snapshot path is not existing
        return True

    for snapshot_dir in snapshot_path.iterdir():
        snapshot_file = snapshot_dir / file_name
        if (
            snapshot_file.exists()
            and snapshot_file.is_symlink()
            and (snapshot_file.resolve() == blob_path)
        ):
            # there is still symlink pointing to the blob file
            # don't delete the blob
            return False

    return True


def file_rm(
    repo_id: str,
    file_name: str,
    revision: str,
) -> bool:
    """Remove target model file."""
    try:
        repo = hf_wrapper.get_repo_info(repo_id)
        rev = hf_wrapper.get_revision_info(repo_id, revision)
        f = hf_wrapper.get_file_info(repo_id, revision, file_name)

        if not repo or not rev or not f:
            logger.info(
                "Repo or file not found: repo=%s, file=%s, rev=%s",
                repo_id,
                file_name,
                revision,
            )
            return False

        # remove snapshot file
        _rm_file(f.file_path, repo.repo_path / "snapshots")

        # remove blob file
        # blob path and file path are the same on windows
        if f.blob_path != f.file_path and (
            _can_delete_blob(
                file_name,
                repo.repo_path / "snapshots",
                f.blob_path,
            )
        ):
            _rm_file(f.blob_path, repo.repo_path / "blobs")

        # if the snapshot dir is not longer existing, it means that the
        # revision is deleted entirely, hence all the refs pointing to
        # the revision should be deleted
        ref_dir = repo.repo_path / "refs"
        if not rev.snapshot_path.exists() and ref_dir.exists():
            ref_files = [ref_dir / ref for ref in rev.refs]
            for ref in ref_files:
                _rm_file(ref, ref_dir)
    except (OSError, ValueError):
        return False

    return True


def repo_rm(repo_id: str, revision: str | None) -> bool:
    """Remove target repo."""
    try:
        repo = hf_wrapper.get_repo_info(repo_id)
        if not repo:
            return True

        for rev in repo.revisions:
            if (
                revision
                and revision not in rev.refs
                and not rev.commit_hash.startswith(revision)
            ):
                continue

            # remove snapshot files
            for f in rev.files:
                file_rm(
                    repo_id,
                    str(f.file_path.relative_to(rev.snapshot_path)),
                    rev.commit_hash,
                )

    except (OSError, ValueError):
        return False

    return True



"""Daemon server."""

import asyncio
import logging
import sys

from aiohttp import web

from hfmc.common.context import HfmcContext
from hfmc.common.api_settings import (
    API_DAEMON_PEERS_ALIVE,
    API_DAEMON_PEERS_CHANGE,
    API_DAEMON_RUNNING,
    API_DAEMON_STOP,
    API_FETCH_FILE_DAEMON,
    API_FETCH_REPO_FILE_LIST,
    API_PEERS_PROBE,
)
from hfmc.daemon.handlers.daemon_handler import (
    alive_peers,
    daemon_running,
    peers_changed,
    stop_daemon,
)
from hfmc.daemon.handlers.fetch_handler import (
    download_file,
    get_repo_file_list,
    search_file,
)
from hfmc.daemon.handlers.peer_handler import pong
from hfmc.daemon.prober import PeerProber

logger = logging.getLogger(__name__)


def _setup_router(app: web.Application) -> None:
    app.router.add_head(API_FETCH_FILE_DAEMON, search_file)
    app.router.add_get(API_FETCH_FILE_DAEMON, download_file)
    app.router.add_get(API_FETCH_REPO_FILE_LIST, get_repo_file_list)

    app.router.add_get(API_PEERS_PROBE, pong)

    app.router.add_get(API_DAEMON_PEERS_ALIVE, alive_peers)
    app.router.add_get(API_DAEMON_STOP, stop_daemon)
    app.router.add_get(API_DAEMON_RUNNING, daemon_running)
    app.router.add_get(API_DAEMON_PEERS_CHANGE, peers_changed)


async def _start() -> None:
    prober = PeerProber(HfmcContext.get_peers())
    HfmcContext.set_peer_prober(prober)
    task = asyncio.create_task(prober.start_probe())  # probe in background
    prober.set_probe_task(task)  # keep strong reference to task

    app = web.Application()
    _setup_router(app)

    runner = web.AppRunner(app)
    await runner.setup()

    all_int_ip = "0.0.0.0"  # noqa: S104
    port = HfmcContext.get_port()
    site = web.TCPSite(runner=runner, host=all_int_ip, port=port)
    await site.start()

    await asyncio.sleep(sys.maxsize)  # keep daemon running


PORT_OCCUPIED = 48


async def start() -> None:
    """Start the daemon server with errors surpressed."""
    try:
        await _start()
    except OSError as e:
        if e.errno == PORT_OCCUPIED:
            port = HfmcContext.get_port()
            logger.info(
                "Target port is already in use. ",
                extra={"port": port},
            )
    except ValueError:
        logger.exception("Daemon start error.")



"""Manager of daemon service."""

from __future__ import annotations

import asyncio
import logging
import platform
import shutil
import signal
import subprocess
from pathlib import Path
from typing import TYPE_CHECKING

from hfmc.client import http_request

if TYPE_CHECKING:
    from argparse import Namespace

logger = logging.getLogger(__name__)

HFMC_EXEC_NAME_GLOBAL = "hfmc"
HFMC_EXEC_NAME_LOCAL = "./main.py"
DELAY_SEC = 2


def _find_executable() -> str | None:
    executable = shutil.which(HFMC_EXEC_NAME_GLOBAL)
    if executable:
        return executable

    main_py = Path(HFMC_EXEC_NAME_LOCAL)
    if main_py.exists() and main_py.is_file():
        return "python " + HFMC_EXEC_NAME_LOCAL

    return None


async def daemon_is_running() -> bool:
    """Check if the HFMC Daemon is running."""
    return await http_request.is_daemon_running()


async def daemon_start(args: Namespace) -> bool:
    """Start the HFMC Daemon in a detached background process."""
    if await daemon_is_running():
        return True

    executable = _find_executable()
    if not executable:
        logger.error("Cannot find HFMC executable.")
        return False

    verbose = "--verbose" if args.verbose else ""
    command = f"{executable} {verbose} daemon start --detach"
    flags = (
        subprocess.CREATE_NO_WINDOW  # type: ignore[attr-defined]
        if (platform.system() == "Windows")
        else 0
    )

    if platform.system() == "Linux":
        # deal with zombie processes on linux
        signal.signal(signal.SIGCHLD, signal.SIG_IGN)

    await asyncio.create_subprocess_shell(
        command,
        stdin=subprocess.DEVNULL,
        stdout=subprocess.DEVNULL,
        stderr=subprocess.DEVNULL,
        creationflags=flags,
    )

    await asyncio.sleep(DELAY_SEC)
    return await daemon_is_running()


async def daemon_stop() -> bool:
    """Stop the HFMC Daemon."""
    if not await daemon_is_running():
        return True

    await http_request.stop_daemon()

    await asyncio.sleep(DELAY_SEC)
    return not await daemon_is_running()



"""Probing the liveness of other peers."""

from __future__ import annotations

import asyncio
import heapq
import logging
from typing import TYPE_CHECKING, List

if TYPE_CHECKING:
    from hfmc.common.peer import Peer

logger = logging.getLogger(__name__)


class PeerProber:
    """Prober for the liveness of other peers."""

    _peers: List[Peer]
    _actives: set
    _updates: set | None
    _probe_heap: List[tuple[int, Peer]]
    _probing: bool
    _probe_task: asyncio.Task[None] | None

    INTERVAL_SEC = 3

    def __init__(self, peers: List[Peer]) -> None:
        """Init PeerProber."""
        self._peers = peers
        self._actives = set()
        self._updates = None
        self._probe_heap = []
        self._probing = False
        self._probe_task = None

    def get_alives(self) -> List[Peer]:
        """Get live peer list."""
        return list(self._actives)

    def update_peers(self, peers: List[Peer]) -> None:
        """Accept a new list of peers to probe."""
        self._updates = set(peers)

    def _reset_peer_heap(self) -> None:
        self._probe_heap = []
        for peer in self._peers:
            heapq.heappush(self._probe_heap, (peer.epoch, peer))

    def _do_update_peers(self) -> None:
        if self._updates is not None:
            peers_removed = set(self._peers) - self._updates
            self._actives = self._actives - peers_removed

            self._peers = list(self._updates)
            self._updates = None

            self._reset_peer_heap()

    async def start_probe(self) -> None:
        """Start probing peers for liveness.

        This function uses asyncio to probe peers for liveness.
        It will wake up every {INTERVAL_SEC} seconds, pop a peer
        from the heap, and then send a ping request to the peer.
        The peer is taken out of the heap until we get a response from
        the peer or the ping request times out.
        After that, the peer is put back into the heap.
        """
        # pylint: disable=import-outside-toplevel
        from hfmc.client.http_request import ping  # resolve cyclic import

        if self._probing:
            return

        self._probing = True
        self._reset_peer_heap()

        if not self._probe_heap:
            logger.debug("No peers configured to probe")

        def probe_cb(task: asyncio.Task[Peer]) -> None:
            try:
                peer = task.result()

                if peer in self._peers:
                    heapq.heappush(self._probe_heap, (peer.epoch, peer))

                if peer.alive and peer in self._peers:
                    self._actives.add(peer)
                else:
                    self._actives.discard(peer)
            except asyncio.exceptions.CancelledError:
                logger.debug("probing is canceled")

        while self._probing:
            await asyncio.sleep(self.INTERVAL_SEC)

            self._do_update_peers()

            if self._probe_heap:
                _, peer = heapq.heappop(self._probe_heap)
                probe = asyncio.create_task(ping(peer))
                probe.add_done_callback(probe_cb)

    def set_probe_task(self, task: asyncio.Task[None]) -> None:
        """Save the coroutine task of probing to avoid gc."""
        self._probe_task = task

    def stop_probe(self) -> None:
        """Stop probing."""
        # TODO: cancel running probe tasks  # noqa: FIX002, TD002, TD003
        self._probing = False
        self._probe_heap = []
        self._actives = set()
        self._probe_task = None



"""Handling repo file list."""

import json
import logging
from pathlib import Path
from typing import List, Optional

from hfmc.common.context import HfmcContext

RepoFileList = List[str]

logger = logging.getLogger(__name__)


def _file_list_local_file(
    repo_id: str,
    revision: str,
) -> Path:
    return HfmcContext.get_repo_files_dir() / repo_id / revision / "files.json"


def load_file_list(
    repo_id: str,
    revision: str,
) -> Optional[RepoFileList]:
    """Load repo file list from local config."""
    path = _file_list_local_file(repo_id, revision)
    if not path.exists():
        return None
    return json.loads(path.read_text())


def save_file_list(repo_id: str, revision: str, files: RepoFileList) -> None:
    """Save repo file list to local config."""
    path = _file_list_local_file(repo_id, revision)
    try:
        if not path.exists():
            path.parent.mkdir(parents=True, exist_ok=True)
            path.touch()
            path.write_text(json.dumps(files))
    except (ValueError, IOError, OSError) as e:
        logger.debug("Error when saving file list.", exc_info=e)



"""Manage peers."""

from __future__ import annotations

from typing import List

from hfmc.client import http_request as request
from hfmc.config import config_manager
from hfmc.config.hfmc_config import HfmcConfigOption, Peer


def _uniq_peers(peers: List[Peer]) -> List[Peer]:
    """Remove duplicate peers."""
    return list(set(peers))


async def add(ip: str, port: int) -> None:
    """Add a peer."""
    peers = config_manager.get_config(HfmcConfigOption.PEERS, List[Peer])
    peers.append(Peer(ip=ip, port=port))
    config_manager.set_config(
        HfmcConfigOption.PEERS,
        _uniq_peers(peers),
        List[Peer],
    )
    await request.notify_peers_change()


async def rm(ip: str, port: int) -> None:
    """Remove a peer."""
    peers = config_manager.get_config(HfmcConfigOption.PEERS, List[Peer])
    peers = [peer for peer in peers if peer.ip != ip or peer.port != port]
    config_manager.set_config(
        HfmcConfigOption.PEERS,
        _uniq_peers(peers),
        List[Peer],
    )
    await request.notify_peers_change()


async def get() -> List[tuple[Peer, bool]]:
    """Get all peers with liveness info."""
    peers = config_manager.get_config(HfmcConfigOption.PEERS, List[Peer])

    # get_alive_peers uses Peer in HfmcContext intead of Peer in HfmcConfig
    alives = {Peer(ip=p.ip, port=p.port) for p in await request.get_alive_peers()}

    return [(peer, peer in alives) for peer in peers]



"""Manager etags for HFMC model files.

This module provides functions to manage etags for Hugging Face model files.
It is needed because huggingface_hub 0.23.0 does not save etags of model files
on Windows.
"""

from __future__ import annotations

import logging
from pathlib import Path

import huggingface_hub as hf  # type: ignore[import-untyped]

from hfmc.common.context import HfmcContext

logger = logging.getLogger(__name__)


def _get_etag_path(repo_id: str, filename: str, revision: str) -> Path | None:
    model_path = hf.try_to_load_from_cache(
        repo_id=repo_id,
        filename=filename,
        revision=revision,
        cache_dir=HfmcContext.get_model_dir_str(),
    )

    # model_path type is (str | Any | None)
    if model_path is None:
        return None

    if not isinstance(model_path, str):
        return None

    rel_path = Path(model_path).relative_to(HfmcContext.get_model_dir())
    return HfmcContext.get_etag_dir() / rel_path


def load_etag(repo_id: str, file_name: str, revision: str) -> str | None:
    """Load etag value from a etag cache file."""
    etag_path = _get_etag_path(repo_id, file_name, revision)

    if not etag_path or not etag_path.exists():
        return None

    return etag_path.read_text().strip()


def save_etag(etag: str, repo_id: str, file_name: str, revision: str) -> None:
    """Save etag value to a etag cache file."""
    etag_path = _get_etag_path(repo_id, file_name, revision)

    if not etag_path:
        logger.debug(
            "Failed to get etag path: repo_id=%s, file_name=%s, revision=%s",
            repo_id,
            file_name,
            revision,
        )
        raise ValueError

    if not etag_path.parent.exists():
        etag_path.parent.mkdir(parents=True, exist_ok=True)

    if not etag_path.exists():
        etag_path.touch()

    etag_path.write_text(etag)



"""Daemon related commands."""

import logging
from argparse import Namespace

from hfmc.daemon import manager as daemon_manager
from hfmc.daemon import server

logger = logging.getLogger(__name__)


async def _daemon_start(args: Namespace) -> None:
    if await daemon_manager.daemon_start(args):
        logger.info("Daemon started.")
    else:
        logger.error("Daemon failed to start.")


async def _daemon_start_detached() -> None:
    await server.start()


async def _daemon_stop() -> None:
    if await daemon_manager.daemon_stop():
        logger.info("Daemon stopped.")
    else:
        logger.error("Daemon failed to stop.")


async def _daemon_status() -> None:
    if await daemon_manager.daemon_is_running():
        logger.info("Daemon is running.")
    else:
        logger.info("Daemon is NOT running.")


async def exec_cmd(args: Namespace) -> None:
    """Execute command."""
    if args.daemon_command == "start" and args.detach:
        await _daemon_start_detached()
    elif args.daemon_command == "start":
        await _daemon_start(args)
    elif args.daemon_command == "stop":
        await _daemon_stop()
    elif args.daemon_command == "status":
        await _daemon_status()
    else:
        raise NotImplementedError



"""Test save and load etag."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, List

import huggingface_hub as hf

from hfmc.common.context import HfmcContext
from hfmc.common.etag import load_etag, save_etag
from hfmc.config.hfmc_config import HfmcConfig

if TYPE_CHECKING:
    import py
    import pytest

ETAG = "1234"
REPO = "test_repo"
FILE = "test_file"
REV = "test_rev"


def test_save_and_load(
    monkeypatch: pytest.MonkeyPatch,
    tmpdir: py.path.local,
) -> None:
    """Test save etag."""

    def mock_module_path(*_: List[Any], **__: dict[str, Any]) -> str:
        return str(HfmcContext.get_model_dir() / REPO / REV / FILE)

    monkeypatch.setattr(
        hf,
        "try_to_load_from_cache",
        mock_module_path,
    )

    conf = HfmcConfig(cache_dir=str(tmpdir))

    HfmcContext.init_with_config(conf)

    save_etag(ETAG, REPO, FILE, REV)
    etag = load_etag(REPO, FILE, REV)

    assert etag == ETAG



"""Test HfmcContext."""

from pytest import fixture

from hfmc.common.context import HfmcContext
from hfmc.common.peer import Peer as ContextPeer
from hfmc.config.hfmc_config import HfmcConfig, Peer


@fixture()
def test_config() -> HfmcConfig:
    """Return a test HfmcConfig."""
    return HfmcConfig(
        cache_dir="test_cache_dir",
        peers=[Peer(ip="127.0.0.1", port=8081)],
        daemon_port=8088,
    )


def test_port(test_config: HfmcConfig) -> None:
    """Test get port."""
    context = HfmcContext.init_with_config(test_config)
    assert context.port == test_config.daemon_port


def test_model_dir(test_config: HfmcConfig) -> None:
    """Test get model dir."""
    context = HfmcContext.init_with_config(test_config)
    assert str(context.model_dir) == "test_cache_dir/models"


def test_etag_dir(test_config: HfmcConfig) -> None:
    """Test get etag dir."""
    context = HfmcContext.init_with_config(test_config)
    assert str(context.etag_dir) == "test_cache_dir/etags"


def test_log_dir(test_config: HfmcConfig) -> None:
    """Test log dir."""
    context = HfmcContext.init_with_config(test_config)
    assert str(context.log_dir) == "test_cache_dir/logs"


def test_get_peers(test_config: HfmcConfig) -> None:
    """Test get peers."""
    context = HfmcContext.init_with_config(test_config)
    assert len(context.peers) == 1

    act_peer = context.peers[0]
    exp_peer = test_config.peers[0]
    assert act_peer.ip == exp_peer.ip
    assert act_peer.port == exp_peer.port
    assert act_peer.alive is False
    assert act_peer.epoch == 0


def test_update_peers() -> None:
    """Test update peers."""
    num_peers = 2
    ip_1, ip_2, ip_3 = "127.0.0.1", "127.0.0.2", "127.0.0.3"
    port_1, port_2, port_3 = 8081, 8082, 8083

    old_conf = HfmcConfig(
        peers=[
            Peer(ip=ip_1, port=port_1),
            Peer(ip=ip_2, port=port_2),
        ],
    )
    old_context = HfmcContext.init_with_config(old_conf)
    assert len(old_context.peers) == num_peers

    old_context.peers[0].alive = True
    old_context.peers[0].epoch = 42

    new_conf = HfmcConfig(
        peers=[
            Peer(ip=ip_1, port=port_1),
            Peer(ip=ip_3, port=port_3),
        ],
    )

    new_peers = HfmcContext.update_peers(new_conf, old_context.peers)

    assert len(new_peers) == num_peers
    assert ContextPeer(ip=ip_1, port=port_1, alive=True, epoch=42) in new_peers
    assert ContextPeer(ip=ip_3, port=port_3, alive=False, epoch=0) in new_peers



"""Handler model and file related requests."""

from __future__ import annotations

import logging
import re
from typing import TYPE_CHECKING, Tuple

import aiofiles
from aiohttp import web

from hfmc.common import hf_wrapper, repo_files
from hfmc.common.etag import load_etag

if TYPE_CHECKING:
    from pathlib import Path

logger = logging.getLogger(__name__)


def _get_file_info(request: web.Request) -> tuple[str, str, str]:
    user = request.match_info["user"]
    model = request.match_info["model"]
    revision = request.match_info["revision"]
    file_name = request.match_info["file_name"]
    repo_id = f"{user}/{model}"
    return repo_id, file_name, revision


async def download_model(_: web.Request) -> web.Response:
    """Download model."""
    raise NotImplementedError


byte_range_re = re.compile(r"bytes=(\d+)-(\d+)?$")


def _get_byte_range(
    request: web.Request,
) -> tuple[int | None, int | None] | None:
    byte_range = request.headers.get("Range")

    if not byte_range or byte_range.strip() == "":
        return None, None

    m = byte_range_re.match(byte_range)
    err_msg = "Invalid byte range: Range=%s"

    if not m:
        logger.error(err_msg, byte_range)
        raise ValueError

    first, last = [int(x) if x else None for x in m.groups()]

    if first is not None and last is not None and last < first:
        logger.error(err_msg, byte_range)
        raise ValueError

    return first, last


async def _file_sender(
    writer: web.StreamResponse,
    file_path: Path,
    file_start: int | None,
    file_end: int | None,
) -> None:
    async with aiofiles.open(file_path, "rb") as f:
        if file_start is not None:
            await f.seek(file_start)

        buf_size = 2**18  # 256 KB buffer size

        while True:
            to_read = min(
                buf_size,
                file_end + 1 - await f.tell() if file_end else buf_size,
            )

            buf = await f.read(to_read)

            if not buf:
                break

            await writer.write(buf)

        await writer.write_eof()


async def download_file(
    request: web.Request,
) -> web.StreamResponse:
    """Download file."""
    br = _get_byte_range(request)
    if br is None:
        return web.Response(status=400)

    file_start, file_end = br
    repo_id, file_name, revision = _get_file_info(request)

    file_info = hf_wrapper.get_file_info(repo_id, revision, file_name)
    if not file_info:
        return web.Response(status=404)

    file_path = file_info.file_path
    if not file_path.exists():
        return web.Response(status=404)

    headers = {"Content-disposition": f"attachment; filename={file_name}"}
    response = web.StreamResponse(headers=headers)
    await response.prepare(request)
    await _file_sender(response, file_path, file_start, file_end)
    return response


async def search_model(_: web.Request) -> web.Response:
    """Search model."""
    raise NotImplementedError


async def search_file(
    request: web.Request,
) -> web.Response:
    """Search file."""
    repo_id, file_name, revision = _get_file_info(request)

    rev_info = hf_wrapper.get_revision_info(repo_id, revision)
    if not rev_info:
        return web.Response(status=404)

    file_info = hf_wrapper.get_file_info(repo_id, revision, file_name)
    if not file_info:
        return web.Response(status=404)

    etag = load_etag(repo_id, file_name, revision)
    return web.Response(
        headers={
            "ETag": etag or "",
            hf_wrapper.COMMIT_HASH_HEADER: rev_info.commit_hash,
            "Content-Length": str(file_info.size_on_disk),
            "Location": str(request.url),
        },
    )


def _get_repo_info(request: web.Request) -> Tuple[str, str]:
    user = request.match_info["user"]
    model = request.match_info["model"]
    revision = request.match_info["revision"]
    repo_id = f"{user}/{model}"
    return repo_id, revision


async def get_repo_file_list(request: web.Request) -> web.Response:
    """Get repo file list."""
    repo_id, revision = _get_repo_info(request)
    files = repo_files.load_file_list(repo_id, revision)
    if not files:
        return web.Response(status=404)
    return web.json_response(files)



"""Commands for authentication."""

from argparse import Namespace
from huggingface_hub import login, logout  # type: ignore[import-untyped]


async def exec_cmd(args: Namespace) -> None:
    """Execute command."""
    if args.auth_command == "login":
        login()
    elif args.auth_command == "logout":
        logout()
    else:
        raise NotImplementedError



"""Logging utils."""

import argparse
import logging
import logging.handlers
import sys

from hfmc.common.context import HfmcContext
from hfmc.utils import args as arg_utils


def _create_stream_handler(level: int) -> logging.Handler:
    """Create a stream handler."""
    handler = logging.StreamHandler(stream=sys.stderr)
    handler.setFormatter(logging.Formatter("%(message)s"))
    handler.setLevel(level)
    return handler


def _create_file_handler(level: int) -> logging.Handler:
    """Create a file handler."""
    log_dir = HfmcContext.get_log_dir()
    log_dir.mkdir(parents=True, exist_ok=True)
    log_file = log_dir / "hfmc.log"

    handler = logging.handlers.RotatingFileHandler(
        log_file,
        maxBytes=10 * 1024 * 1024,  # 10MB
        backupCount=5,
    )

    log_format = "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
    handler.setFormatter(logging.Formatter(log_format))
    handler.setLevel(level)

    return handler


def setup_logging(args: argparse.Namespace) -> None:
    """Set up logging."""
    # configure root logger
    level = arg_utils.get_logging_level(args)

    if arg_utils.is_detached_daemon(args):
        handler = _create_file_handler(level)
    else:
        handler = _create_stream_handler(level)

    root_logger = logging.getLogger()
    root_logger.addHandler(handler)
    root_logger.setLevel(level)

    # suppress lib's info log
    logging.getLogger("asyncio").setLevel(logging.WARNING)
    logging.getLogger("aiohttp").setLevel(logging.WARNING)
    logging.getLogger("urllib3").setLevel(logging.WARNING)
    logging.getLogger("filelock").setLevel(logging.WARNING)
    logging.getLogger("huggingface_hub").setLevel(logging.WARNING)



"""Entrypoint of HFMC."""

import asyncio
import logging
from argparse import Namespace

from hfmc.client import model_cmd, peer_cmd, uninstall_cmd
from hfmc.common.context import HfmcContext
from hfmc.config import conf_cmd, config_manager
from hfmc.daemon import daemon_cmd
from hfmc.utils import auth_cmd, logging as logging_utils
from hfmc.utils.args import arg_parser

logger = logging.getLogger(__name__)


async def _exec_cmd(args: Namespace) -> None:
    if args.command == "daemon":
        exec_cmd = daemon_cmd.exec_cmd
    elif args.command == "peer":
        exec_cmd = peer_cmd.exec_cmd
    elif args.command == "model":
        exec_cmd = model_cmd.exec_cmd
    elif args.command == "conf":
        exec_cmd = conf_cmd.exec_cmd
    elif args.command == "auth":
        exec_cmd = auth_cmd.exec_cmd
    elif args.command == "uninstall":
        exec_cmd = uninstall_cmd.exec_cmd
    else:
        raise NotImplementedError

    await exec_cmd(args)


async def _async_main() -> None:
    config = config_manager.load_config()
    HfmcContext.init_with_config(config)

    args = arg_parser()
    logging_utils.setup_logging(args)

    await _exec_cmd(args)


def main() -> None:
    """Entrypoint of HFMC."""
    try:
        asyncio.run(_async_main())
    except (
        KeyboardInterrupt,
        asyncio.exceptions.CancelledError,
    ):
        # ignore interrupt and cancel errors as they are handled by daemon
        logger.info("Shutting down HFMC.")



"""Model management related commands."""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING, List

from prettytable import PrettyTable

from hfmc.client import model_controller

if TYPE_CHECKING:
    from argparse import Namespace

    from hfmc.client.model_controller import FileInfo, RepoInfo

logger = logging.getLogger(__name__)


def _tablize(names: List[str], rows: List[List[str]]) -> None:
    table = PrettyTable()
    table.field_names = names
    table.add_rows(rows)
    logger.info(table)


def _tablize_files(files: List[FileInfo]) -> None:
    if not files:
        logger.info("No files found.")
    else:
        names = ["REFS", "COMMIT", "FILE", "SIZE", "PATH"]
        rows = [
            [
                ",".join(f.refs),
                f.commit_8,
                str(f.file_name),
                str(f.size_on_disk_str),
                str(f.file_path),
            ]
            for f in files
        ]
        _tablize(names, rows)


def _tablize_repos(repos: List[RepoInfo]) -> None:
    if not repos:
        logger.info("No repos found.")
    else:
        names = ["REPO ID", "SIZE", "NB FILES", "LOCAL PATH"]
        rows = [
            [
                r.repo_id,
                f"{r.size_str:>12}",
                str(r.nb_files),
                str(
                    r.repo_path,
                ),
            ]
            for r in repos
        ]
        _tablize(names, rows)


def _ls(args: Namespace) -> None:
    if args.repo:
        files = model_controller.file_list(args.repo)
        _tablize_files(files)
    else:
        repos = model_controller.repo_list()
        _tablize_repos(repos)


async def _add(args: Namespace) -> None:
    if args.file is None and args.revision == "main":
        msg = (
            "In order to keep repo version integrity, when add a repo,"
            "You must specify the commit hash (i.e. 8775f753) with -v option."
        )
        logger.info(msg)
        return

    if args.file:
        target = f"File {args.repo}/{args.file}"
        success = await model_controller.file_add(
            args.repo,
            args.file,
            args.revision,
        )
    else:
        target = f"Model {args.repo}"
        success = await model_controller.repo_add(
            args.repo,
            args.revision,
        )

    if success:
        logger.info("%s added.", target)
    else:
        logger.info("%s failed to add.", target)


def _rm(args: Namespace) -> None:
    if args.file:
        if not args.revision:
            logger.info("Remove file failed, must specify the revision!")
            return

        target = "File"
        success = model_controller.file_rm(
            args.repo,
            args.file,
            args.revision,
        )
    else:
        target = "Model"
        success = model_controller.repo_rm(
            args.repo,
            args.revision,
        )

    if success:
        logger.info("%s remove is done.", target)
    else:
        logger.info("%s failed to remove.", target)


async def _search(args: Namespace) -> None:
    if args.file:
        target = "file"
        peers = await model_controller.file_search(
            args.repo,
            args.file,
            args.revision,
        )
    else:
        target = "model"
        peers = await model_controller.repo_search()

    if peers:
        logger.info(
            "Peers that have target %s:\n[%s]",
            target,
            ",".join(
                [f"{p.ip}:{p.port}" for p in peers],
            ),
        )
    else:
        logger.info("NO peer has target %s.", target)


async def exec_cmd(args: Namespace) -> None:
    """Execute command."""
    if args.model_command == "ls":
        _ls(args)
    elif args.model_command == "add":
        await _add(args)
    elif args.model_command == "rm":
        _rm(args)
    elif args.model_command == "search":
        await _search(args)
    else:
        raise NotImplementedError



"""A wrapper of huggingface_hub api."""

from __future__ import annotations
import logging
from typing import List

import huggingface_hub as hf  # type: ignore[import-untyped]
from huggingface_hub.hf_api import HfApi  # type: ignore[import-untyped]

from hfmc.common.context import HfmcContext
from hfmc.common.repo_files import RepoFileList

COMMIT_HASH_HEADER = hf.constants.HUGGINGFACE_HEADER_X_REPO_COMMIT
logger = logging.getLogger(__name__)


def get_cache_info() -> hf.HFCacheInfo:
    """Get cache info."""
    return hf.scan_cache_dir(HfmcContext.get_model_dir_str())


def get_repo_info(repo_id: str) -> hf.CachedRepoInfo | None:
    """Get repo info by repo_id."""
    cache_info = get_cache_info()
    for repo in cache_info.repos:
        if repo.repo_id == repo_id:
            return repo
    return None


def get_revision_info(
    repo_id: str,
    revision: str,
) -> hf.CachedRevisionInfo | None:
    """Get revision info by revision."""
    repo_info = get_repo_info(repo_id)

    if repo_info is None:
        return None

    for rev in repo_info.revisions:
        if revision in rev.refs or rev.commit_hash.startswith(revision):
            return rev

    return None


def get_file_info(
    repo_id: str,
    revision: str,
    filename: str,
) -> hf.CachedFileInfo | None:
    """Get file info by filename."""
    rev_info = get_revision_info(repo_id, revision)

    if rev_info is None:
        return None

    for f in rev_info.files:
        if rev_info.snapshot_path / filename == f.file_path:
            return f

    return None


def get_repo_file_list(
    endpoint: str,
    repo_id: str,
    revision: str,
) -> RepoFileList | None:
    """Load repo struct."""
    fs = hf.HfFileSystem(endpoint=endpoint)
    repo = f"{repo_id}@{revision}/"
    path = f"hf://{repo}"
    try:
        beg = len(repo)
        return [f[beg:] for f in fs.find(path)]
    except (ValueError, OSError, IOError):
        logger.debug(
            "Cannot load repo file list for %s, %s, %s",
            endpoint,
            repo_id,
            revision,
        )
        return None


def verify_revision(
    repo_id: str,
    revision: str,
    endpoints: List[str],
) -> str | None:
    """Verify if revision is valid."""
    # verify with local cache
    rev_info = get_revision_info(repo_id, revision)
    if rev_info:
        return rev_info.commit_hash

    # verify with remote endpoints
    for endpoint in endpoints:
        api = HfApi(endpoint=endpoint)
        try:
            model = api.model_info(repo_id, revision=revision)
            if model and model.sha:
                return model.sha
        except (OSError, IOError, ValueError):
            continue

    return None



"""Handle requests to daemon."""

from dataclasses import asdict

from aiohttp import web
from aiohttp.web_runner import GracefulExit

from hfmc.common.context import HfmcContext
from hfmc.config import config_manager


async def alive_peers(_: web.Request) -> web.Response:
    """Find alive peers."""
    alives = HfmcContext.get_peer_prober().get_alives()
    return web.json_response([asdict(peer) for peer in alives])


async def peers_changed(_: web.Request) -> web.Response:
    """Update peers."""
    config = config_manager.load_config()
    new_peers = HfmcContext.update_peers(config, HfmcContext.get_peers())
    HfmcContext.get_peer_prober().update_peers(new_peers)
    return web.Response()


async def stop_daemon(request: web.Request) -> None:
    """Stop the daemon."""
    HfmcContext.get_peer_prober().stop_probe()

    resp = web.Response()
    await resp.prepare(request)
    await resp.write_eof()

    raise GracefulExit


async def daemon_running(_: web.Request) -> web.Response:
    """Check if daemon is running."""
    return web.Response()



"""Test cases for the daemon client."""

import pytest

from hfmc.client.http_request import ping
from hfmc.common.peer import Peer


@pytest.mark.asyncio()
async def test_peers_probe() -> None:
    """Test probe a live peer."""
    peer = Peer("127.0.0.2", 8080)
    await ping(peer)



"""HFMC Configuration Class."""

from __future__ import annotations

from enum import Enum
from pathlib import Path
from typing import List

from pydantic import BaseModel, Field

CONFIG_DIR = Path.home() / ".hfmc"
CONFIG_FILE = CONFIG_DIR / "config.yaml"

DEFAULT_CACHE_DIR = Path.home() / ".cache" / "hfmc"
DEFAULT_DAEMON_PORT = 9090


class Peer(BaseModel):
    """Peer definition for HFMC."""

    ip: str = Field(exclude=False, frozen=True)
    port: int = Field(exclude=False, frozen=True)

    def __lt__(self, other: object) -> bool:
        """Return True if self is less than other."""
        if isinstance(other, Peer):
            return self.ip < other.ip or (
                self.ip == other.ip and self.port < other.port
            )
        return NotImplemented

    def __eq__(self, other: object) -> bool:
        """Return True if self is equal to other."""
        if isinstance(other, Peer):
            return self.ip == other.ip and self.port == other.port
        return NotImplemented

    def __hash__(self) -> int:
        """Return the hash value of the Peer."""
        return hash((self.ip, self.port))


class HfmcConfigOption(str, Enum):
    """HFMC configuration options."""

    CACHE: str = "cache_dir"
    PORT: str = "daemon_port"
    PEERS: str = "peers"


class HfmcConfig(BaseModel):
    """Data class for HFMC directory configuration."""

    cache_dir: str = Field(
        description="Directory for storing cache files",
        default=str(DEFAULT_CACHE_DIR),
    )

    peers: List[Peer] = Field(
        description="List of peers",
        default_factory=list,
    )

    daemon_port: int = Field(
        description="Port for the daemon",
        default=DEFAULT_DAEMON_PORT,
    )



"""Peer related commands."""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING, List

from hfmc.client import peer_controller

if TYPE_CHECKING:
    from argparse import Namespace

    from hfmc.config.hfmc_config import Peer

logger = logging.getLogger(__name__)


def _print_peer_list(peers: List[tuple[Peer, bool]]) -> None:
    """Print peer list."""
    for peer, alive in peers:
        peer_name = f"{peer.ip}:{peer.port}"
        peer_stat = "alive" if alive else ""
        peer_str = f"{peer_name}\t{peer_stat}"
        logger.info(peer_str)


async def exec_cmd(args: Namespace) -> None:
    """Execute command."""
    if args.peer_command == "add":
        await peer_controller.add(args.ip, args.port)
    elif args.peer_command == "rm":
        await peer_controller.rm(args.ip, args.port)
    elif args.peer_command == "ls":
        peers = await peer_controller.get()
        _print_peer_list(peers)
    else:
        raise NotImplementedError



"""Test the config_manager module."""

from typing import cast

import hfmc.config.config_manager as manager
from hfmc.config.hfmc_config import (
    CONFIG_FILE,
    DEFAULT_CACHE_DIR,
    DEFAULT_DAEMON_PORT,
    HfmcConfig,
    Peer,
)
from hfmc.utils.yaml import yaml_load


def test_init() -> None:
    """Test default initialization."""
    if CONFIG_FILE.exists():
        CONFIG_FILE.unlink()

    manager.init_config()

    conf = cast(dict, yaml_load(CONFIG_FILE))

    assert conf == {
        "cache_dir": str(DEFAULT_CACHE_DIR),
        "peers": [],
        "daemon_port": DEFAULT_DAEMON_PORT,
    }


def test_save_and_load() -> None:
    """Test saving and loading configuration."""
    custom = {
        "cache_dir": "custom_cache_dir",
        "peers": [{"ip": "127.0.0.1", "port": 8080}],
        "daemon_port": 8080,
    }

    peers = [Peer(ip=p["ip"], port=p["port"]) for p in custom["peers"]]
    cache_dir = custom["cache_dir"]
    daemon_port = custom["daemon_port"]

    conf = HfmcConfig(
        cache_dir=cache_dir,
        peers=peers,
        daemon_port=daemon_port,
    )

    manager.save_config(conf)
    saved = cast(dict, yaml_load(CONFIG_FILE))
    assert saved == custom

    loaded = manager.load_config()
    assert loaded == conf


def test_change_config() -> None:
    """Test set, reset, and get configs."""
    port_1 = 1234
    port_2 = 4321

    ip = "192.168.0.1"
    cache_dir = "new_cache_dir"

    manager.set_config(manager.ConfOpt.CACHE, cache_dir, str)
    manager.set_config(manager.ConfOpt.PORT, port_1, int)
    manager.set_config(
        manager.ConfOpt.PEERS,
        [Peer(ip=ip, port=4321)],
        list,
    )

    assert manager.get_config(manager.ConfOpt.CACHE, str) == cache_dir
    assert manager.get_config(manager.ConfOpt.PORT, int) == port_1
    assert manager.get_config(manager.ConfOpt.PEERS, list) == [
        Peer(ip=ip, port=port_2),
    ]

    manager.reset_config(manager.ConfOpt.CACHE, str)
    manager.reset_config(manager.ConfOpt.PORT, int)
    manager.reset_config(manager.ConfOpt.PEERS, list)

    assert manager.get_config(manager.ConfOpt.CACHE, str) == str(
        DEFAULT_CACHE_DIR,
    )
    assert manager.get_config(manager.ConfOpt.PORT, int) == DEFAULT_DAEMON_PORT
    assert manager.get_config(manager.ConfOpt.PEERS, list) == []



"""Utils for args parsing and usage."""

from __future__ import annotations

import argparse
import logging
from argparse import Namespace

from hfmc.common.context import HfmcContext


def is_detached_daemon(args: Namespace) -> bool:
    """Check if HFMC is running as a detached daemon."""
    return args.command == "daemon" and args.daemon_command == "start" and args.detach


def get_logging_level(args: Namespace) -> int:
    """Get logging level from args."""
    if args.verbose:
        return logging.DEBUG
    return logging.INFO


# pylint: disable=too-many-locals,too-many-statements
def arg_parser() -> Namespace:  # noqa: PLR0915
    """Parse args."""
    df_port = HfmcContext.get_port()

    parser = argparse.ArgumentParser(prog="hfmc")
    parser.add_argument("-v", "--verbose", action="store_true")
    subparsers = parser.add_subparsers(dest="command", required=True)

    # hfmc daemon ...
    daemon_parser = subparsers.add_parser("daemon")
    daemon_subparsers = daemon_parser.add_subparsers(
        dest="daemon_command",
        required=True,
    )
    # hfmc daemon start ...
    daemon_start_parser = daemon_subparsers.add_parser("start")
    daemon_start_parser.add_argument("-d", "--detach", action="store_true")
    # hfmc daemon stop
    daemon_subparsers.add_parser("stop")
    # hfmc daemon status
    daemon_subparsers.add_parser("status")

    # hfmc peer ...
    peer_parser = subparsers.add_parser("peer")
    peer_subparsers = peer_parser.add_subparsers(
        dest="peer_command",
        required=True,
    )
    # hfmc peer add ...
    peer_add_parser = peer_subparsers.add_parser("add")
    peer_add_parser.add_argument("ip")
    peer_add_parser.add_argument("-p", "--port", type=int, default=df_port)
    # hfmc peer rm ...
    peer_rm_parser = peer_subparsers.add_parser("rm")
    peer_rm_parser.add_argument("ip")
    peer_rm_parser.add_argument("-p", "--port", type=int, default=df_port)
    # hfmc peer ls ...
    peer_subparsers.add_parser("ls")

    # hfmc model ...
    model_parser = subparsers.add_parser("model")
    model_subparsers = model_parser.add_subparsers(
        dest="model_command",
        required=True,
    )
    # hfmc model ls ...
    model_ls_parser = model_subparsers.add_parser("ls")
    model_ls_parser.add_argument("-r", "--repo")
    # hfmc model add ...
    model_add_parser = model_subparsers.add_parser("add")
    model_add_parser.add_argument("-r", "--repo", required=True)
    model_add_parser.add_argument("-f", "--file")
    model_add_parser.add_argument("-v", "--revision", default="main")
    # hfmc model rm ...
    model_rm_parser = model_subparsers.add_parser("rm")
    model_rm_parser.add_argument("-r", "--repo", required=True)
    model_rm_parser.add_argument("-f", "--file")
    model_rm_parser.add_argument("-v", "--revision")
    # hfmc model search ...
    model_search_parser = model_subparsers.add_parser("search")
    model_search_parser.add_argument("-r", "--repo", required=True)
    model_search_parser.add_argument("-f", "--file")
    model_search_parser.add_argument("-v", "--revision", default="main")

    # hfmc conf ...
    conf_parser = subparsers.add_parser("conf")
    conf_subparsers = conf_parser.add_subparsers(
        dest="conf_command",
        required=True,
    )
    # hfmc conf cache ...
    conf_cache_parser = conf_subparsers.add_parser("cache")
    conf_cache_subparsers = conf_cache_parser.add_subparsers(
        dest="conf_cache_command",
        required=True,
    )
    conf_cache_set_parser = conf_cache_subparsers.add_parser("set")
    conf_cache_set_parser.add_argument("path")
    conf_cache_subparsers.add_parser("get")
    conf_cache_subparsers.add_parser("reset")
    # hfmc conf port ...
    conf_port_parser = conf_subparsers.add_parser("port")
    conf_port_subparsers = conf_port_parser.add_subparsers(
        dest="conf_port_command",
        required=True,
    )
    conf_port_set_subparser = conf_port_subparsers.add_parser("set")
    conf_port_set_subparser.add_argument("port", type=int)
    conf_port_subparsers.add_parser("get")
    conf_port_subparsers.add_parser("reset")
    # hfmc conf show
    conf_subparsers.add_parser("show")

    # hfmc auth ...
    auth_parser = subparsers.add_parser("auth")
    auth_subparsers = auth_parser.add_subparsers(
        dest="auth_command",
        required=True,
    )
    # hfmc auth login
    auth_subparsers.add_parser("login")
    # hfmc auth logout
    auth_subparsers.add_parser("logout")

    # hfmc uninstall
    subparsers.add_parser("uninstall")

    return parser.parse_args()



"""Uninstall HFMC by removing related directories."""

import logging
import shutil
from argparse import Namespace

from hfmc.config import config_manager
from hfmc.config.hfmc_config import CONFIG_DIR, HfmcConfigOption
from hfmc.daemon import manager as daemon_manager

logger = logging.getLogger(__name__)


async def _uninstall() -> None:
    if await daemon_manager.daemon_is_running():
        logger.info("Stop daemon first by executing 'hfmc daemon stop'.")
        return

    warning = (
        "WARNING: 'Uninstall' will delete all hfmc data on disk, "
        "and it's not recoverable!."
    )
    logging.info(warning)

    confirm = input("Please enter 'Y/y' to confirm uninstall: ")

    if confirm not in ["Y", "y"]:
        logger.info("Cancel uninstall.")
        return

    cache_dir = config_manager.get_config(HfmcConfigOption.CACHE, str)
    home_dir = str(CONFIG_DIR)
    to_rm = [cache_dir, home_dir]

    for d in to_rm:
        shutil.rmtree(d, ignore_errors=True)

    logger.info("HFMC is uninstalled.")


async def exec_cmd(args: Namespace) -> None:
    """Execute command."""
    if args.command == "uninstall":
        await _uninstall()
    else:
        raise NotImplementedError
